""" Utility functions """
import traceback

######################################################################
# # Logging initialisation
######################################################################
import logging
try:
    from cloghandler import ConcurrentRotatingFileHandler as RFHandler
except ImportError:
    from logging.handlers import RotatingFileHandler as RFHandler

def initLogger():
    # create handler
    handler = RFHandler('logging.log', maxBytes=1000000, backupCount=10)
    handler.setLevel(logging.DEBUG)
    # create formatter
    formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(process)-5d [%(module)s.%(funcName)s:%(lineno)d] %(message)s")
    # add formatter to handler
    handler.setFormatter(formatter)
    # add handler to logger
    log = logging.getLogger('')
    log.setLevel(logging.DEBUG)
    log.addHandler(handler)
    logging.info('Initialisation')

######################################################################
# # XML data copy functions
######################################################################
from lxml import etree
from copy import deepcopy

def getValue(infoSet, xpath, module=None, nss={}):
    value = None
    if module != None:
        location = '(' + str(module) + ')'
    else:
        location = ''
    try:
        value_list = infoSet.xpath(xpath, namespaces=nss)
        if len(value_list) > 0:
            if type(value_list[0]) == etree._Element:
                value = value_list[0].text
                logging.debug('%s=%s %s' % (xpath, value, location))
            elif type(value_list[0]) == etree._ElementStringResult:
                value = value_list[0]
                logging.debug('%s=%s %s' % (xpath, value, location))
            else:
                logging.warning('Unsupported type: %s for %s %s' % (type(value_list[0], xpath, location)))
        else:
            logging.warning('Not found: %s %s' % (xpath, location))
    except:
        exception = str(traceback.format_exc())
        logging.warning('Could not get: %s %s: %s' % (xpath, location, exception))
    return value

def setValue(infoSet, xpath, value, module=None, nss={}):
    if module != None:
        location = '(' + str(module) + ')'
    else:
        location = ''
    if value != None:
        try:
            value_list = infoSet.xpath(xpath, namespaces=nss)
            if len(value_list) > 0:
                if type(value_list[0]) == etree._Element:
                    value_list[0].text = value
                    logging.debug('%s=%s %s' % (xpath, value, location))
                else:
                    logging.warning('Unsupported type: %s for %s %s' % (type(value_list[0], xpath, location)))
            else:
                logging.warning('Could not set: %s=%s %s' % (xpath, value, location))
        except:
            exception = str(traceback.format_exc())
            logging.warning('Could not set: %s=%s %s\n%s' % (xpath, value, location, exception))
        
def setAttrValue(infoSet, xpath, name, value, module=None, nss={}):
    if module != None:
        location = '(' + str(module) + ')'
    else:
        location = ''
    if value != None and value != '':
        try:
            value_list = infoSet.xpath(xpath, namespaces=nss)
            if len(value_list) > 0:
                if type(value_list[0]) == etree._Element:
                    value_list[0].set(name, value)
                    logging.debug('%s/@%s=%s %s' % (xpath, name, value, location))
                else:
                    logging.warning('Unsupported type: %s for %s %s' % (type(value_list[0], xpath, location)))
            else:
                logging.warning('Not found: %s %s' % (xpath, location))
        except:
            exception = str(traceback.format_exc())
            logging.warning('Could not set: %s/@%s=%s %s\n%s' % (xpath, name, value, location, exception))

def duplicateStructure(infoSet, xpath, nss={}):
    try:
        value_list = infoSet.xpath(xpath, namespaces=nss)
        if len(value_list) > 0:
            if type(value_list[0]) == etree._Element:
                copy = deepcopy(value_list[0])
                parent = value_list[0].getparent()
                index = parent.index(value_list[0])
                parent.insert(index + 1, copy)
                logging.debug('Inserted a duplicate %s in %s' % (copy.tag, parent.tag))
            else:
                logging.warning('Unsupported type: %s for %s %s' % (type(value_list[0], xpath)))
        else:
            logging.warning('Not found: %s %s' % (xpath))
    except:
        exception = str(traceback.format_exc())
        logging.warning('Could not duplicate: %s %s: %s' % (xpath, exception))
    return

######################################################################
# # XML Schema registration and validation functions
######################################################################

# Compiled W3C XML Schemas, Schematron Schemas, XSLT scripts
CompiledXSD = {}
CompiledSCH = {}
CompiledXSLT = {}
# Compiled pre-parsed XML documents for use as templates
PreparsedXML = {}

def registerXSD(ns, file):
    """
    Load and cache an XML schema for namespace C{ns} from C{file}

    @param ns: Namespace
    @type ns: C{str}
    @param file: Path to XML Schema file (.xsd file)
    @type file: C{str}
    """
    try:
        if ns in CompiledXSD:
            logging.debug('Schema %s is already registered!' % ns)
        else:
            logging.info('Registering schema %s from %s' % (ns, file))
            xmlschemadoc = etree.parse(file)
            xsd = etree.XMLSchema(xmlschemadoc)
            CompiledXSD[ns] = xsd
    except:
        exception = str(traceback.format_exc())
        logging.error("Exception while trying to register XSD:\n%s" % exception)

def initSchematron():
    # Compile the Schematron compiler
    try:
        registerXSLT('iso_svrl_for_xslt1.xsl', '../stylesheets/iso_svrl_for_xslt1.xsl')
    except:
        logging.error('Error registering Schematron compiler')

def registerSCH(id, file, phase=None):
    """
    Compiles and registers a Schematron schema associated with a specific I{id} from I{file} 
    @param id: identifier that the Schematron schema will be associated with
    @param file: file to load the Schematron schema from
    @param phase: Schematon phase (default is default in Schematron definition set)
    """
    try:
        if id in CompiledSCH:
            logging.debug('Schematron rule set %s is already registered!' % str(id))
        else:
            if phase == None:
                logging.info('Registering Schematron rule set %s from %s (default phase)' % (str(id), file))
            else: 
                logging.info('Registering Schematron rule set %s from %s (phase \'%s\')' % (str(id), file, phase))
            schtransform = CompiledXSLT['iso_svrl_for_xslt1.xsl']
            schdoc = etree.parse(file)

            if phase == None:
                sch = schtransform(schdoc)
            else:
                sch = schtransform(schdoc, phase=unicode("\'" + phase + "\'"))

            string = str(sch)
            schtransformdoc = etree.fromstring(string)
            CompiledSCH[id] = etree.XSLT(schtransformdoc)

    except:
        exception = str(traceback.format_exc())
        logging.error("Exception while trying to register Schematron file:\n%s" % exception)

def validateXMLDocumentX(ns, infoset, xsdvalidate=True, schvalidate=True):
    """
    Extended validation of an I{infoset} against the XML schema and/or Schematron associated with namespace I{ns}

    @param ns: namespace of the schemas that the I{infoset} should be associated with
    @param xsdvalidate: if an XML schema is registered under I{ns}, whether or not the I{infoset} should be
    validated against it
    @param schvalidate: if a Schematron ruleset is registered under I{ns}, whether or not the I{infoset}
    should be validated against it
    @return: A L{DocumentValidationReport<DocumentValidationReport>} class instance
    
    """

    xsdvalidated = False
    xsdisvalid = False
    schvalidated = False
    schisvalid = False

    # First Pass: XML Schema validation
    if xsdvalidate and ns in CompiledXSD:
        xsd = CompiledXSD[ns]
        if xsd.validate(infoset):
            logging.info('XSD validation OK ')
            xsdvalidated = True
            xsdisvalid = True
            xsddetails = ''
        else:
            xsdvalidated = True
            xsdisvalid = False
            xsddetails = xsd.error_log.last_error
            logging.warning('XSD validation FAILED:\n%s\n', xsddetails)

    else:
        logging.info('No XSD registered for \"%s\"', ns)
        xsdvalidated = False
        xsdisvalid = False
        xsddetails = None

    # Second Pass: Schematron validation
    if schvalidate and ns in CompiledSCH:
        sch = CompiledSCH[ns]
        report = sch(infoset)
        failedassertions = report.xpath('//svrl:failed-assert/svrl:text', namespaces={'svrl': 'http://purl.oclc.org/dsdl/svrl'})
        if len(failedassertions) == 0:
            logging.info('Schematron validation OK ')
            schisvalid = True
            schdetails = None
        else:
            schisvalid = False
            schdetails = report
            result = ''
            for assertion in failedassertions:
                result = result + unicode(assertion.text) + '\n'
            logging.warning('Schematron validation FAILED:\n%s\n', result)
        schvalidated = True
    else:
        logging.info('No Schematron rules registered for \"%s\"', ns)
        schvalidated = False
        schisvalid = False
        schdetails = None

    valid = (not xsdvalidated or xsdisvalid) and (not schvalidated or schisvalid)
    return DocumentValidationReport(valid, xsdvalidated, xsdisvalid, xsddetails, schvalidated, schisvalid, schdetails)

class DocumentValidationReport():
    def __init__(self, valid, xsdvalidated, xsdisvalid, xsddetails, schvalidated, schisvalid, schdetails):
        """
        This is Separate class that represents a report of validating an XML document. It indicates if the
        document was valid, which types of validation were performed, and detailed reports from those
        validations.  At the moment both XSD and Schematron validation is supported.

        @param self: DocumentValidationReport instance
        @param valid: valid
        @param xsdvalidated: xsdvalidated
        @param xsdisvalid: xsdisvalid
        @param xsddetails: xsddetails
        @param schvalidated: schvalidated
        @param schisvalid:
        @param schdetails:
        """
        self.valid = valid
        
        self.xsdvalidated = xsdvalidated
        self.xsdvalid = xsdisvalid
        self.xsddetails = xsddetails

        self.schvalidated = schvalidated
        self.schvalid = schisvalid
        self.schdetails = schdetails

######################################################################
# # XML template registration
######################################################################

def registerPreParsedXML(id, template):
    """
    Parse an XML document from a template file into an in memory Etree for performance
    
    @param id: an identifier under which the etree is stored 
    @type id: C{str}
    @param template: the template to load from
    @type template: C{str}
    """
    try:
        if id in PreparsedXML:
            logging.debug('XML template %s is already registered!' % id)
        else:
            logging.info('Registering XML template %s from %s' % (id, template))
            tree = etree.parse(template)
            PreparsedXML[id] = tree
    except:
        exception = str(traceback.format_exc())
        logging.error("Exception while trying to register XML template %s from %s: %s" % (id, template, exception))

def retrievePreParsedXML(id):
    """
    Retrieve the Etree registered under an id. 
    
    @param id: an identifier under which the etree is stored 
    @type id: C{str}
    """
    logging.debug('Retrieving XML template %s' % (id))
    tree = PreparsedXML[id]
    return deepcopy(tree)

######################################################################
# # XSLT stylesheet registration and transformation functions
######################################################################

def registerXSLT(id, file):
    """
    Load XSLT Stylesheet for namespace C{id} from C{file} and cache compiled version

    @param id: Identifier
    @type id: C{str}
    @param file: Path to XSLT stylesheet file (.XSLT file)
    @type file: C{str}
    """
    try:
        if id in CompiledXSLT:
            logging.debug('XSLT stylesheet %s is already registered!' % id)
        else:
            logging.info('Registering XSLT stylesheet %s from %s' % (id, file))
            xsltxml = etree.parse(file)
            transform = etree.XSLT(xsltxml)
            CompiledXSLT[id] = transform
    except:
        exception = str(traceback.format_exc())
        logging.error("Exception while trying to register XSLT:\n%s" % exception)

def transform(id, infoset, **kwargs):
    """
    Transforms an I{infoset} using a transformation registered with identifier I{id}
    @param id: identifier under which the stylesheet is registerd
    @type id: C{str}
    @param infoset: infoset to transform
    @type infoset: infoset
    @return: transformed document
    """
    logging.info("Transforming document using stylesheet %s", id) 
    transform = CompiledXSLT[id]
    result = transform(infoset, **kwargs)
    logging.info("Transformation done") 
    return result

def xsltParam(string):
    """
    Auxiliary function to prepare parameters to be passed to lxml compiled XSLT stylesheets.
    """
    return unicode("\'" + unicode(string) + "\'")

######################################################################
# # file I/O functions
######################################################################

def serialize(infoSet, file):
    """
    Serialize the C{infoSet} to an encoded string representation of its XML tree
    and write it to the C{file}
    
    @param infoSet: The InfoSet to serialize
    @type infoSet: L{lxml.etree.Element}
    @param file: The file that will be created
    @type file: C{str}
    """
    fd = open(file, 'w')
    fd.write(etree.tostring(infoSet, xml_declaration=True, encoding='UTF-8', pretty_print=True))
    fd.close()
